/* -*- tab-width:8; coding:utf-8 -*-
 * 
 * -- This is code taken from libdvdcss's ioctl.c -- rs@mytum.de
 *
 * Original authors of ioctl.c: 
 *          Markus Kuespert <ltlBeBoy@beosmail.com>
 *          Sam Hocevar <sam@zoy.org>
 *          Jon Lech Johansen <jon-vl@nanocrew.net>
 *          Håkan Hjort <d95hjort@dtek.chalmers.se>
 *          Eugenio Jarosiewicz <ej0@cise.ufl.edu>
 *          David Siebörger <drs-videolan@rucus.ru.ac.za>
 *          Alex Strelnikov <lelik@os2.ru>
 *          Gildas Bazin <gbazin@netcourrier.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#if defined (__SVR4) && defined (__sun)

#include <dlfcn.h>
#include <string.h>
#include <unistd.h>

#include "solaris_uscsi.h"

/*****************************************************************************
 * SolarisInitUSCSI: initialize a USCSICMD structure for the Solaris kernel
 *****************************************************************************
 * This function initializes a Solaris userspace scsi command structure for
 * future use, either a read command or a write command.
 *****************************************************************************/
void SolarisInitUSCSI( struct uscsi_cmd *p_sc, int i_type )
{
    union scsi_cdb *rs_cdb;
    memset( p_sc->uscsi_cdb, 0, sizeof( union scsi_cdb ) );
    memset( p_sc->uscsi_bufaddr, 0, p_sc->uscsi_buflen );

    switch( i_type )
    {
        case GPCMD_SEND_KEY:
            p_sc->uscsi_flags = USCSI_ISOLATE | USCSI_WRITE;
            break;

        case GPCMD_READ_DVD_STRUCTURE:
        case GPCMD_REPORT_KEY:
            p_sc->uscsi_flags = USCSI_ISOLATE | USCSI_READ;
            break;
    }

    rs_cdb = (union scsi_cdb *)p_sc->uscsi_cdb;

    rs_cdb->scc_cmd = i_type;

    rs_cdb->cdb_opaque[ 8 ] = (p_sc->uscsi_buflen >> 8) & 0xff;
    rs_cdb->cdb_opaque[ 9 ] =  p_sc->uscsi_buflen       & 0xff;
    p_sc->uscsi_cdblen = 12;

    USCSI_TIMEOUT( p_sc, 15 );
}

/*****************************************************************************
 * SolarisSendUSCSI: send a USCSICMD structure to the Solaris kernel
 * for execution
 *****************************************************************************
 * When available, this function uses the function smedia_uscsi_cmd()
 * from Solaris' libsmedia library (Solaris 9 or newer) to execute the
 * USCSI command.  smedia_uscsi_cmd() allows USCSI commands for
 * non-root users on removable media devices on Solaris 9; sending the
 * USCSI command directly to the device using the USCSICMD ioctl fails
 * with an EPERM error on Solaris 9.
 *
 * The code will fall back to the USCSICMD ioctl method, when
 * libsmedia.so is not available or does not export the
 * smedia_uscsi_cmd() function (on Solaris releases upto and including
 * Solaris 8). Fortunatelly, on these old releases non-root users are
 * allowed to perform USCSICMD ioctls on removable media devices.
 *****************************************************************************/
int SolarisSendUSCSI( int i_fd, struct uscsi_cmd *p_sc )
{
    void *p_handle;

    /* We use static variables to keep track of the libsmedia symbols, which
     * is harmless even in a multithreaded program because the library and
     * its symbols will always be mapped at the same address. */
    static int b_tried = 0;
    static int b_have_sm = 0;

    typedef void * (*t_get_handle) ( int32_t );
    typedef int (*t_uscsi_cmd) ( void *, struct uscsi_cmd * );
    typedef int (*t_release_handle) ( void * );
    static t_get_handle p_get_handle;
    static t_uscsi_cmd p_uscsi_cmd;
    static t_release_handle p_release_handle;

    if( !b_tried )
    {
        void *p_lib;

        p_lib = dlopen( "libsmedia.so", RTLD_NOW );
        if( p_lib )
        {
            p_get_handle = (t_get_handle)dlsym( 
		p_lib, "smedia_get_handle" );
            p_uscsi_cmd = (t_uscsi_cmd)dlsym( 
		p_lib, "smedia_uscsi_cmd" );
            p_release_handle = (t_release_handle)dlsym(
		p_lib, "smedia_release_handle" );

            if( p_get_handle && p_uscsi_cmd && p_release_handle )
            {
                b_have_sm = 1;
            }
            else
            {
                dlclose( p_lib );
            }
        }

        b_tried = 1;
    }

    if( b_have_sm && (p_handle = p_get_handle(i_fd)) )
    {
        int i_ret = p_uscsi_cmd( p_handle, p_sc );
        p_release_handle( p_handle );
        return i_ret;
    }

    return ioctl( i_fd, USCSICMD, p_sc );
}

#endif
